#define DEBUG
/*
 * @(#)AbacusT.c
 *
 * Copyright 2009  David A. Bagley, bagleyd@tux.org
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Used tables from How to Learn Lee's Abacus by Lee Kai-chen.
 */

#include "AbacusP.h"

#define CHAR_TO_DIGIT(c) ((c >= 'A') ? c - 'A' + 10 : c - '0')
#define DIGIT_TO_CHAR(d) ((d >= 10) ? (char) ('A' + d - 10) : (char) ('0' + d))
#define IS_DIGIT(c) ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'J'))

/* ignores decimal point and multiply signs, etc, etc */
static void simpleParser(AbacusWidget w, char * buffer,
		char* aString, char* bString, char *op)
{
	unsigned int i;
	int numberCount = 0;
	Boolean decimal = False;
	int digitCount = 0;
	int decimalPosition = w->abacus.decimalPosition;
	
	if (w->abacus.decks[BOTTOM].piece != 0)
		decimalPosition--;
	if (w->abacus.decimalPosition >= w->abacus.shiftPercent &&
			w->abacus.decks[BOTTOM].piecePercent != 0)
		decimalPosition--;
	aString[0] = '\0';
	bString[0] = '\0';
	*op = ' ';
	for (i = 0; i < strlen(buffer); i++) {
		if (buffer[i] == 'q' || buffer[i] == 'Q') {
			*op = buffer[i];
			return;
		}
		if (buffer[i] == '+' || buffer[i] == '-' ||
				buffer[i] == '*' || buffer[i] == '/' ||
				buffer[i] == 'v' || buffer[i] == 'u') {
			if (!decimal) {
				if (numberCount == 0)
					(void) sprintf(aString, "%s%c", aString, '.');
				else
					(void) sprintf(bString, "%s%c", bString, '.');
			}
			for (; digitCount < decimalPosition; digitCount++) {
				if (numberCount == 0)
					(void) sprintf(aString, "%s%c", aString, '0');
				else
					(void) sprintf(bString, "%s%c", bString, '0');
			}
			if (*op != ' ')
				return;
			*op = buffer[i];
			numberCount++;
			digitCount = 0;
			decimal = False;
		} else if (buffer[i] == '.' && !decimal) {
			decimal = True;
			if (numberCount == 0)
				(void) sprintf(aString, "%s%c", aString, '.');
			else
				(void) sprintf(bString, "%s%c", bString, '.');
		} else if (IS_DIGIT(buffer[i])) {
			if (!decimal || digitCount < decimalPosition) {
				if (numberCount == 0) {
					(void) sprintf(aString, "%s%c", aString, buffer[i]);
				} else {
					(void) sprintf(bString, "%s%c", bString, buffer[i]);
				}
			}
			if (decimal) {
				digitCount++;
			}
		}
	}
	if (!decimal) {
		if (numberCount == 0) {
			(void) sprintf(aString, "%s%c", aString, '.');
		} else {
			(void) sprintf(bString, "%s%c", bString, '.');
		}
	}
	for (; digitCount < decimalPosition; digitCount++) {
		if (numberCount == 0)
			(void) sprintf(aString, "%s%c", aString, '0');
		else
			(void) sprintf(bString, "%s%c", bString, '0');
	}
}
/* More generic */
#if 1
/* Taken from addition tables in "How to Learn Lee's Abacus */
/* Author does not really discuss carry and says to work left */ 
/* to right.  I do not know how to have that work smoothly so */ 
/* it works right to left. */
static void digitAdd(int a, int b, int carryIn,
		int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carryOut)
{
	int newB = b + carryIn;
	int fractBase;
	int modB, modAns, ans;
	int divA, modDivAns, divAns;

	if (bottomNumber > base / 2) /* topFactor may not be set, Russian */
		fractBase = base;
	else
		fractBase = topFactor;
	modB = newB % fractBase;
	ans = a + newB;
	modAns = ans % fractBase;
	divA = a / fractBase;
	modDivAns = (ans % base) / fractBase;
	divAns = ans / base;
	if (modAns - modB >= 0)
		*lower = modB;
	else
		*lower = modB - fractBase;
	*upper = modDivAns - divA;
	*carryOut = divAns;
#ifdef DEBUG
	(void) printf("add:\t%d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, carryIn,
		*lower, *upper, *carryOut);
#endif
}

/* Taken from subtraction tables in "How to Learn Lee's Abacus,
   made a little more generic. */
static void digitSubtract(int a, int b, int carryIn,
		int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carryOut)
{
	int newB = b - carryIn;
	int fractBase;
	int modA, modB, ans;
	int divA, modDivAns, divAns;

	if (bottomNumber > base / 2) /* topFactor may not be set, Russian */
		fractBase = base;
	else
		fractBase = topFactor;
	modA = a % fractBase;
	modB = newB % fractBase;
	ans = a - newB;
	divA = a / fractBase;
	modDivAns = ((ans + base) % base) / fractBase;
	divAns = (ans - base + 1) / base;
	if (modA - modB >= 0)
		*lower = -modB;
	else
		*lower = fractBase - modB;
	*upper = modDivAns - divA;
	*carryOut = divAns;
#ifdef DEBUG
	(void) printf("sub:\t%d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, carryIn,
		*lower, *upper, *carryOut);
#endif
}

#else
/* less generic */
static void digitAdd(int a, int b, int carryIn,
		int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carryOut)
{
	int newB = b + carryIn;
	int fractBase;

	if (bottomNumber > base / 2) /* topFactor may not be set */
		fractBase = base;
	else
		fractBase = topFactor;
	*lower = 0;
	*upper = 0;
	*carryOut = 0;
	if (newB < fractBase) {
		if (a < fractBase) {
			if ((a + newB) < fractBase) { /* table 1 */
				*lower = newB;
			} else { /* table 2 */
				*lower = newB - fractBase;
				if (bottomNumber > base / 2)
					*carryOut = 1;
				else
					*upper = 1;
			}
		} else {
			if ((a + newB) < base) { /* table 1 */
				*lower = newB;
			} else { /* table 3 */
				*lower = newB - fractBase;
				*upper = -1;
				*carryOut = 1;
			}
		}
	} else {
		if ((a + newB) < base) { /* table 1 */
			*lower = newB - fractBase;
			*upper = 1;
		} else if (a >= fractBase && a + newB < base + fractBase) { /* table 4 */
			*lower = newB - fractBase;
			*upper = -1;
			*carryOut = 1;
		} else { /* table 3 */
			*lower = newB - base;
			*carryOut = 1;
		}
	}
#ifdef  DEBUG
	(void) printf("add %d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, carryIn,
		*lower, *upper, *carryOut);
#endif
}

/* Taken from subtraction tables in "How to Learn Lee's Abacus */
static void digitSubtract(int a, int b, int carryIn,
		int base, int topFactor, int bottomNumber,
		int* lower, int* upper, int* carryOut)
{
	int newB = b - carryIn;
	int fractBase; /* Kludge but may not be a bad fix here */
	if (bottomNumber > base / 2) /* topFactor may not be set */
		fractBase = base;
	else
		fractBase = topFactor;
	*lower = 0;
	*upper = 0;
	*carryOut = 0; /* if used is negative, i.e. a borrow */
	if (newB <= fractBase) {
		if (a < fractBase) {
			if ((a - newB) >= 0) { /* table 1 */
				*lower = -newB;
			} else { /* table 3 */
				*lower = fractBase - newB;
				if (bottomNumber <= base / 2)
					*upper = 1;
				*carryOut = -1;
			}
		} else {
			if ((a - newB) >= fractBase) { /* table 2, 3 */
				*lower = -newB;
			} else { /* table 1 */
				*lower = fractBase - newB;
				*upper = -1;
			}
		}
	} else {
		if ((a - newB) >= 0) { /* table 1 */
			*lower = fractBase - newB;
			*upper = -1;
		} else if (a < fractBase && (a - newB) >= -fractBase) { /* table 4 */
			*lower = fractBase - newB;
			*upper = 1;
			*carryOut = -1;
		} else { /* table 3 */
			*lower = base - newB;
			*carryOut = -1;
		}
	}
#ifdef DEBUG
	(void) printf("sub %d\t%d\t%d\t|\t%d\t%d\t%d\n", a, b, carryIn,
		*lower, *upper, *carryOut);
#endif
}
#endif

#ifdef OLD_DEBUG
static void testTable()
{
	int i, j, carryIn, k, lower, upper, carryOut;
#if 1
	int base = DEFAULT_BASE;
	int factor = 5;
#else
	int base = 16;
	int factor = 4;
#endif
	int c = 1;
	/*int c = 2;*/

	for (k = 0 ; k < c; k++) {
		for (j = 0 ; j < base; j++) {
			for (i = 0 ; i < base; i++) {
				carryIn = k;
#if 0
				digitAdd(i, j, carryIn,
					base, factor, factor,
					&lower, &upper, &carryOut);
#else
				digitSubtract(i, j, carryIn,
					base, factor, factor,
					&lower, &upper, &carryOut);
#endif
				(void) printf("%d\t%d\t%d\t|\t%d\t%d\t%d\n",
					i, j, carryIn, lower, upper, carryOut);
			}
		}
	}
}
#endif

static void drawLineText(AbacusWidget w, char* line, int i)
{
#ifdef WINVER
#ifdef DEMO_FRAMED
	(void) SetTextColor(w->core.hDC, w->abacusDemo.foregroundGC);
	(void) SetBkMode(w->core.hDC, TRANSPARENT);
	(void) TextOut(w->core.hDC,
		1,
		w->core.height + w->abacusDemo.fontHeight * i - 1,
		line, (signed) strlen(line));
#else
	drawTeachText(line, i);
#endif
#else
	setAbacusText(w, ACTION_TEACH_LINE, line, i);
#endif
}

static int getDecimalStringPosition(char * string)
{
	int i = 0;

	while (string[i] != '\0') {
		if (string[i] == '.')
			return i;
		i++;
	}
	return i;
}

/* Appends decimal where necessary for consistency */
static void decimalSafe(char * aString, char * bString)
{
	int aDecimal = -1, bDecimal = -1;

	aDecimal = getDecimalStringPosition(aString);
	if (aDecimal == (int) strlen(aString)) {
		(void) sprintf(aString, "%s%c", aString, '.');
	}
	bDecimal = getDecimalStringPosition(bString);
	if (bDecimal == (int) strlen(bString)) {
		(void) sprintf(bString, "%s%c", bString, '.');
	}
}

static int decimalPlace(char * string, int pos)
{
	int i = getDecimalStringPosition(string);

	if (i == (int) strlen(string) || i >= pos)
		return i - pos - 1;
	else
		return i - pos;
}

static int decimalOffset(char * string, int place)
{
	int i = getDecimalStringPosition(string);

	if (place < 0)
		return i - place;
	else
		return i - place - 1;
}

/* contract StringBuffer to remove leading and trailing 0's */
static void contractStringBuffer(char * string)
{
	int offset = getDecimalStringPosition(string);
	int i;

	for (i = 0; i < offset - 1; i++) {
		if (string[0] == '0')
			strcpy(string, &(string[1]));
		else
			break;
	}
	if (offset < (int) strlen(string))
		for (i = (int) strlen(string) - 1; i > 1; i--) {
			if (string[i] == '0')
				string[i] = '\0';
			else
				break;
		}
}

/* expand StringBuffer to fit decimal place */
static void expandStringBuffer(char * string, int place)
{
	int offset = getDecimalStringPosition(string);
	int prependOffset = place - offset + 1;
	int appendOffset = offset + 1 - (int) strlen(string) - place;
	int i, len;

	len = (int) strlen(string);
	if (place > 0 && prependOffset > 0) {
		for (i = len; i >= 0; i--) {
			string[i + prependOffset] = string[i];
		}
		for (i = 0; i < prependOffset; i++) {
			string[i] = '0';
		}
	} else if (place < 0 && appendOffset > 0) {
		for (i = 0; i < appendOffset; i++) {
			string[len + i] = '0';
		}
		string[len + appendOffset] = '\0';
	}
}

static void setStringBuffer(AbacusWidget w, char * string,
		char op, int place, int lower, int upper)
{
	int offset, digit;
	int topFactor = w->abacus.decks[TOP].factor;
	int bottomNumber = w->abacus.decks[BOTTOM].number;

	if (bottomNumber <= DEFAULT_BASE / 2)
		digit = lower + upper * topFactor;
	else
		digit = lower;
	expandStringBuffer(string, place);
	offset = decimalOffset(string, place);
	if (op == '+') {
		string[offset] = DIGIT_TO_CHAR((digit +
			CHAR_TO_DIGIT(string[offset])) % w->abacus.base);
	} else /* if (op == '-') */ {
		string[offset] = DIGIT_TO_CHAR((digit + w->abacus.base -
			CHAR_TO_DIGIT(string[offset])) % w->abacus.base);
	}
}

/* Appends 0's where necessary to adding is easier */
static void addSafe(char * aString, char * bString)
{
	int aDecimal = -1, bDecimal = -1;
	int aCount, bCount, i;

	aDecimal = getDecimalStringPosition(aString);
	bDecimal = getDecimalStringPosition(bString);
	aCount = (int) strlen(aString) - aDecimal;
	bCount = (int) strlen(bString) - bDecimal;
	char temp[512];

	if (aCount > bCount) {
		for (i = bCount; i < aCount; i++)
			(void) sprintf(bString, "%s%c", bString, '0');
	} else {
		for (i = aCount; i < bCount; i++)
			(void) sprintf(aString, "%s%c", aString, '0');
	}
	if (aDecimal > bDecimal) {
		for (i = bDecimal; i < aDecimal; i++) {
			(void) strcpy(temp, bString);
			(void) sprintf(bString, "%c%s", '0', temp);
		}
	} else {
		for (i = aDecimal; i < bDecimal; i++) {
			(void) strcpy(temp, aString);
			(void) sprintf(aString, "%c%s", '0', temp);
		}
	}
}

static int nextRail(AbacusWidget w, int step, int aux, Boolean rightToLeft,
		char * string)
{
	int n, rail;
	int count = (int) strlen(string) - 1;
	int decimal = 0;
	int shiftPercent = w->abacus.shiftPercent;
	Boolean piece = checkPiece(w);
	Boolean piecePercent = checkPiecePercent(w);
	Boolean subdeck = checkSubdeck(w, 3);
	int decimalPosition = (int) strlen(string) - 1 -
		getDecimalStringPosition(string);

	n = step;
	if (rightToLeft) {
		if (w->abacus.carryStep != 0)
			n += (w->abacus.carryStep / 2);
		if (n >= decimalPosition -
			((piece) ? 1 : 0) -
				((piecePercent) ? 1 : 0) -
				((subdeck) ? 2 : 0)) {
			decimal++;
		}
		n += decimal;
		rail = n - decimalPosition;
		if (subdeck) {
			rail += 2;
		} else {
			if (rail >= -1 + ((piecePercent) ? 1 : 0) &&
					piece)
				rail++;
			if (rail >= -shiftPercent - 2 &&
					piecePercent)
				rail++;
			if (n >= decimalPosition) {
				rail--;
			}
		}
	} else {
		if (w->abacus.carryStep != 0)
			n += -(w->abacus.carryStep / 2);
		if (n >= count - decimalPosition +
				((piece) ? 1 : 0) +
				((piecePercent) ? 1 : 0) -
				((subdeck) ? 2 : 0)) {
			decimal++;
		}
		rail = count - n - decimalPosition - 1;
		n += decimal;
		if (subdeck) {
			rail += 3;
		} else {
			if (piece && n <= count - decimalPosition +
					((piecePercent) ? 1 : 0))
				rail++;
			if (n <= count - decimalPosition +
					shiftPercent + 2 &&
					piecePercent)
				rail++;
			if (n <= count - decimalPosition +
					shiftPercent + 2 &&
					piecePercent)
				rail++;
		}
	}
#ifdef DEBUG
	(void) printf("nextRail: %d\n", rail);
#endif
	return rail;
}

static int nextCharPosition(AbacusWidget w, int step, int carryStep, int aux,
		Boolean rightToLeft, char * string)
{
	int n, a;
	int count = (int) strlen(string) - 1;
	int decimal = 0;
	Boolean piece = checkPiece(w);
	Boolean piecePercent = checkPiecePercent(w);
	Boolean subdeck = checkSubdeck(w, 3);
	int decimalPosition = (int) strlen(string) - 1 -
		getDecimalStringPosition(string);

	n = step;
	if (rightToLeft) {
		if (carryStep != 0)
			n += (carryStep / 2);
		if (n >= decimalPosition -
				((piece) ? 1 : 0) -
				((piecePercent) ? 1 : 0) -
				((subdeck) ? 2 : 0)) {
			decimal++;
		}
		n += decimal;
		if (n > count)
			a = -1;
		else
			a = count - n;
	} else {
		if (carryStep != 0)
			n += -(carryStep / 2);
		if (n >= count - decimalPosition +
				((piece) ? 1 : 0) +
				((piecePercent) ? 1 : 0) -
				((subdeck) ? 2 : 0)) {
			decimal++;
		}
		n += decimal;
		if (n < 0)
			a = -1;
		else
			a = n;
	}
#ifdef DEBUG
	(void) printf("nextCharPosition: %d at step %d in %s\n", a, step, string);
#endif
	return a;
}

static int nextChar(char * string, int pos)
{
	int digit;

	if (pos < 0 || pos >= (int) strlen(string)) {
		digit = 0;
	} else {
		digit = CHAR_TO_DIGIT(string[pos]);
	}
#ifdef DEBUG
	(void) printf("nextChar: %d at %d in %s\n", digit, pos, string);
#endif
	return digit;
}

static int addSteps(char * string)
{
	/* decimal included */
	return (int) strlen(string) - 1;
}

static int multSteps(char * aString, char * bString)
{
	/* decimal included */
	return ((int) strlen(aString) - 1) * ((int) strlen(bString) - 1);
}

static int addMultSteps(char * aString, char * bString)
{
	/* 2 digits per multiplication */
	return 2 * multSteps(aString, bString);
}

static void headDigits(AbacusWidget w, char * aString, char * bString,
		char * substring)
{
	int aLen = (int) strlen(aString) - 1; /* decimal point */
	int bLen = (int) strlen(bString) - 1;
	long a, b;

	if (aLen < 0 || bLen < 0 || aLen < bLen) {
		substring[0] = '\0';
		return;
	}
	(void) strncpy(substring, aString, (unsigned) bLen);
	substring[bLen + 1] = '\0';
	a = convertToDecimal(w->abacus.base, substring);
	b = convertToDecimal(w->abacus.base, bString);
	if (b - a > 0) {
		if (aLen == bLen) {
			substring[0] = '\0';
			return;
		}
		(void) strncpy(substring, aString, (unsigned) bLen + 1);
		substring[bLen + 2] = '\0';
		a = convertToDecimal(w->abacus.base, substring);
	}
}

static int offsetDigits(char * string, char * hString)
{
	int i;

	for (i = (int) strlen(hString); i < (int) strlen(string); i++) {
		if (string[i] == '.')
			return i - (int) strlen(hString);
	}
	return i - (int) strlen(hString);
}

static Boolean nextPositionSum(AbacusWidget w, char op,
		int *rPosition, int *rDigit, int *bDigit)
{
	int n = (w->abacus.step - 2) / 2; /* 2 step display */
	int max = addSteps(w->abacus.aString);
	int topFactor = w->abacus.decks[TOP].factor;
	int bottomNumber = w->abacus.decks[BOTTOM].number;
	Boolean rightToLeft = w->abacus.rightToLeftAdd;
	int rPos, bPos, place;

	/* rString can expand with carries. */
	/* bString does not change, so bPos will be predictable. */
	bPos = nextCharPosition(w, n, w->abacus.carryStep, 0,
		rightToLeft, w->abacus.bString);
	place = decimalPlace(w->abacus.bString, bPos);
	rPos = decimalOffset(w->abacus.rString, place);
	*rDigit = nextChar(w->abacus.rString, rPos);
	if (w->abacus.carryStep == 0)
		*bDigit = nextChar(w->abacus.bString, bPos);
	else
		*bDigit = 1;
	if (!rightToLeft)
		w->abacus.carry[w->abacus.state] = 0;
	if (op == '+')
		digitAdd(*rDigit, *bDigit,
			w->abacus.carry[w->abacus.state],
			w->abacus.base, topFactor, bottomNumber,
			&(w->abacus.lower), &(w->abacus.upper),
			&(w->abacus.carry[w->abacus.state]));
	else
		digitSubtract(*rDigit, *bDigit,
			w->abacus.carry[w->abacus.state],
			w->abacus.base, topFactor, bottomNumber,
			&(w->abacus.lower), &(w->abacus.upper),
			&(w->abacus.carry[w->abacus.state]));
	*rPosition = nextRail(w, n, 0, rightToLeft, w->abacus.bString);
	return (n >= max - 1);
}

static Boolean nextPositionProduct(AbacusWidget w,
		int *rPosition, int *aDigit, int *bDigit)
{
	int n = (w->abacus.step - 2) / 2; /* 2 step display */
	int max = addMultSteps(w->abacus.aString, w->abacus.bString);
	int topFactor = w->abacus.decks[TOP].factor;
	int bottomNumber = w->abacus.decks[BOTTOM].number;
	int bCount = (int) strlen(w->abacus.bString) - 1;
	Boolean rightToLeft = w->abacus.rightToLeftAdd;
	int aOffset, bOffset, rOffset;
	int aPlace, bPlace, rPlace, rDigit;
	int bValue = 0;

	aOffset = nextCharPosition(w,
		(n / 2) / bCount, /* 2 digits result for each multiplication */
		0, /* not place for carry */
		1,     /* aux 1 */
		True,  /* "a" side always starts on right */
		w->abacus.aString);
	aPlace = decimalPlace(w->abacus.aString, aOffset);
	*aDigit = nextChar(w->abacus.aString, aOffset);
	bOffset = nextCharPosition(w,
		(n / 2) % bCount, /* 2 digits result for each multiplication */
		0, /* not place for carry */
		2,     /* aux 2 */
		w->abacus.rightToLeftMult, /* this can vary */
		w->abacus.bString);
	bPlace = decimalPlace(w->abacus.bString, bOffset);
	*bDigit = nextChar(w->abacus.bString, bOffset);
	rPlace = ((n % 2 == 0 && !rightToLeft) ||
		(n % 2 == 1 && rightToLeft)) ? 1 : 0;
	if (w->abacus.carryStep != 0) {
		rPlace += w->abacus.carryStep / 2;
		*bDigit = 1;
	}
	rPlace += aPlace + bPlace;
	expandStringBuffer(w->abacus.rString, rPlace);
	rOffset = decimalOffset(w->abacus.rString, rPlace);
	rDigit = nextChar(w->abacus.rString, rOffset);
	if (w->abacus.carryStep == 0) {
		/* 2 digits * 2 step display */
		if (((n % 2 == 0) && w->abacus.rightToLeftAdd) ||
		    ((n % 2 == 1) && !w->abacus.rightToLeftAdd)) {
			bValue = (*aDigit * *bDigit) % w->abacus.base;
		} else {
			bValue = (*aDigit * *bDigit) / w->abacus.base;
		}
	} else {
		*aDigit = 0;
		*bDigit = 0;
		bValue = 1;
	}
	if (!rightToLeft)
		w->abacus.carry[w->abacus.state] = 0;
	digitAdd(rDigit, bValue,
		w->abacus.carry[w->abacus.state],
		w->abacus.base, topFactor, bottomNumber,
		&(w->abacus.lower), &(w->abacus.upper),
		&(w->abacus.carry[w->abacus.state]));
	*rPosition = rPlace;
	if (w->abacus.carry[w->abacus.state] != 0 && n % 2 == 1 &&
			w->abacus.rightToLeftAdd)
		return True; /* Do not want to forget carry */
	return (n >= max - 1);
}

/* Plan:
	Take smallest head(rString)-> hrString where aString < hrString
	hrString div aString = factor -> add to bString with offset
	aString * bString = substring
	Appr offset for substring
	subtract offset substring from rString -> new rstring
	repeat while integer
*/
static Boolean nextPositionDivision(AbacusWidget w,
		int *rPosition, int *aDigit, int *bDigit)
{
#ifdef DEBUG
	char hString[1024];
	long a, b, divi, subvalue;
	char * substring = hString;

	headDigits(w, w->abacus.aString, w->abacus.bString, substring);
	(void) printf("%s %s %s\n", w->abacus.aString, w->abacus.bString,
		hString);
	(void) printf("%d\n", offsetDigits(w->abacus.aString, hString));
	a = convertToDecimal(w->abacus.base, w->abacus.aString);
	b = convertToDecimal(w->abacus.base, w->abacus.bString);
	divi = a / b;
	subvalue = b * divi;
	(void) printf("%ld %ld\n", divi, subvalue);
#endif
	return False; /* division hard to know when done */
}


/* A little tricky as this is reentrant */
void teachStep(AbacusWidget w, char * buffer, int aux)
{
	char buffer1[120] = "";
	char buffer2[120] = "";
	int base = w->abacus.base;
	int bottomNumber = w->abacus.decks[BOTTOM].number;
	double a, b;

	if (w->abacus.step == 0) {
#ifdef OLD_DEBUG
		testTable();
#endif
		drawLineText(w, "Enter calculation a+b, a-b, or a*b, where a positive and result positive.", 0);
		drawLineText(w, "Press enter to go through calculation steps.", 1);
#ifndef WINVER
		if (w->abacus.aString)
			free(w->abacus.aString);
		if (w->abacus.bString)
			free(w->abacus.bString);
		if (w->abacus.rString)
			free(w->abacus.rString);
		if ((w->abacus.aString = (char *) malloc(sizeof (char) *
				w->abacus.numDigits + 2)) &&
		   (w->abacus.bString = (char *) malloc(sizeof (char) *
				w->abacus.numDigits + 2)) &&
		   (w->abacus.rString = (char *) malloc(sizeof (char) *
				w->abacus.numDigits + 2)))
#endif
			w->abacus.step++;
	} else if (w->abacus.step == 1) {
		simpleParser(w, buffer, w->abacus.aString, w->abacus.bString,
			&(w->abacus.op));
#ifdef DEBUG
		(void) printf("buffer = %s %c %s\n",
			w->abacus.aString, w->abacus.op, w->abacus.bString);
#endif
		if (w->abacus.op == 'q' || w->abacus.op == 'Q') {
			return;
		}
		if (w->abacus.anomaly != 0) {
			addBackAnomaly(w->abacus.aString, w->abacus.anomaly,
				w->abacus.shiftAnomaly,
				base);
		}
		if (w->abacus.anomalySq != 0) {
			addBackAnomaly(w->abacus.aString, w->abacus.anomalySq,
				w->abacus.shiftAnomaly + w->abacus.shiftAnomalySq,
				base);
		}
		if (checkSubdeck(w, 3)) {
			zeroFractionalPart(w->abacus.aString);
			zeroFractionalPart(w->abacus.bString);
		}
		contractStringBuffer(w->abacus.aString);
		contractStringBuffer(w->abacus.bString);
		if (w->abacus.op == '+' || w->abacus.op == '-') {
			convertStringToAbacus(w, w->abacus.aString, aux);
			convertStringToAbacus(w, "0.0", 1);
			convertStringToAbacus(w, "0.0", 2);
			a = convertToDecimal(base, w->abacus.aString);
			b = convertToDecimal(base, w->abacus.bString);
			if (w->abacus.op == '-' && b > a) {
				/* Revisit this, it should be allowed, but
				   rails probably need to be respected. */
				(void) sprintf(buffer1, "Subtraction overflow %s - %s",
					w->abacus.aString, w->abacus.bString);
				drawLineText(w, buffer1, 0);
				return;
			}
			(void) sprintf(buffer1, "%sing %s %c %s",
				((w->abacus.op == '+') ? "Add" : "Subtract"),
				w->abacus.aString, w->abacus.op,
				w->abacus.bString);
			drawLineText(w, buffer1, 0);
			decimalSafe(w->abacus.aString, w->abacus.bString);
			addSafe(w->abacus.aString, w->abacus.bString);
			(void) strcpy(w->abacus.rString, w->abacus.aString);
			w->abacus.step++;
		} else if (w->abacus.op == '*') {
			convertStringToAbacus(w, "0.0", 0);
			convertStringToAbacus(w, w->abacus.aString, 1);
			a = convertToDecimal(base, w->abacus.aString);
			convertStringToAbacus(w, w->abacus.bString, 2);
			b = convertToDecimal(base, w->abacus.bString);
			/* TODO? add getLee() */
			(void) sprintf(buffer1, "Multiplying %s %c %s",
				w->abacus.aString, w->abacus.op,
				w->abacus.bString);
			drawLineText(w, buffer1, 0);
			decimalSafe(w->abacus.aString, w->abacus.bString);
			(void) sprintf(w->abacus.rString, "0.");
			w->abacus.step++;
		} else if (w->abacus.op == '/') {
			convertStringToAbacus(w, "0.0", 2);
			convertStringToAbacus(w, w->abacus.aString, 0);
			a = convertToDecimal(base, w->abacus.aString);
			convertStringToAbacus(w, w->abacus.bString, 1);
			b = convertToDecimal(base, w->abacus.bString);
			if (b == 0.0) {
				(void) sprintf(buffer1, "Division overflow %s / %s",
					w->abacus.aString, w->abacus.bString);
				drawLineText(w, buffer1, 0);
			}
			/* TODO? add getLee() */
			(void) sprintf(buffer1, "Dividing %s %c %s",
				w->abacus.aString, w->abacus.op,
				w->abacus.bString);
			drawLineText(w, buffer1, 0);
			w->abacus.step++;
		} else if (w->abacus.op == 'v' || w->abacus.op == 'u') {
			convertStringToAbacus(w, w->abacus.aString, aux);
			convertStringToAbacus(w, "0.0", 1);
			convertStringToAbacus(w, "0.0", 2);
			a = convertToDecimal(base, w->abacus.aString);
			(void) sprintf(buffer1, "%s root of %s",
				((w->abacus.op == 'v') ? "Square" : "Cube"),
				w->abacus.aString);
			drawLineText(w, buffer1, 0);
			w->abacus.step++;
		}
#ifdef DEBUG
			(void) printf("op buffer: %s %c %s %d %d\n",
				w->abacus.aString, w->abacus.op,
                                w->abacus.bString, w->abacus.rightToLeftAdd,
                                w->abacus.rightToLeftMult);
#endif
		w->abacus.carry[0] = 0;
		w->abacus.carry[1] = 0;
	} else if ((w->abacus.carryStep != 0 && w->abacus.carryStep % 2 == 0 &&
			w->abacus.carryStep >= 2) ||
			(w->abacus.carryStep == 0 && w->abacus.step % 2 == 0 &&
			w->abacus.step >= 2)) {
		int rPosition = 0;
		int aDigit = 0, bDigit = 0;
		Boolean done = False;

		w->abacus.state = 0;
		if (w->abacus.op == '+' || w->abacus.op == '-') {
			done = nextPositionSum(w, w->abacus.op,
				&rPosition, &aDigit, &bDigit);
		} else if (w->abacus.op == '*') {
			done = nextPositionProduct(w,
				&rPosition, &aDigit, &bDigit);
			(void) sprintf(buffer1, "Multiplying %s %c %s",
				w->abacus.aString, w->abacus.op,
				w->abacus.bString);
			if (w->abacus.carryStep == 0) {
				(void) sprintf(buffer1, "%s ... %d * %d = %d",
					buffer1, aDigit, bDigit, (aDigit * bDigit));
			} else {
				(void) sprintf(buffer1, "%s ... carrying %d",
					buffer1, 1);
			}
			drawLineText(w, buffer1, 0);
		} else if (w->abacus.op == '/') {
			(void) sprintf(buffer1, "Dividing %s %c %s, still in development",
				w->abacus.aString, w->abacus.op,
				w->abacus.bString);
		} else if (w->abacus.op == 'v' || w->abacus.op == 'u') {
			drawLineText(w, "I have not started teaching abacus how to do this.", 0);
		}
		(void) sprintf(buffer2, "For rail %d", rPosition);
		if (w->abacus.lower == 0 && w->abacus.upper == 0 &&
				w->abacus.carry[w->abacus.state] == 0) {
			(void) sprintf(buffer2, "%s, do nothing", buffer2);
			w->abacus.step++;
		} else {
			done = False;
			if (w->abacus.lower != 0) {
			  (void) sprintf(buffer2, "%s, %s %d",
			    buffer2,
			    ((w->abacus.lower < 0) ? "take off" : "put on"),
			    ABS(w->abacus.lower));
			  if (bottomNumber <= DEFAULT_BASE / 2)
			    (void) sprintf(buffer2, "%s %s lower deck",
			      buffer2,
			      ((w->abacus.lower < 0) ? "from" : "to"));
			}
			if (w->abacus.upper != 0) {
			  (void) sprintf(buffer2,
			    "%s, %s %d %s upper deck",
			    buffer2,
			    ((w->abacus.upper < 0) ? "take off" : "put on"),
			    ABS(w->abacus.upper),
			    ((w->abacus.upper < 0) ? "from" : "to"));
			}
			if (w->abacus.carry[w->abacus.state] != 0) {
			    (void) sprintf(buffer2, "%s, %s %d (on next move)",
			      buffer2,
			      ((w->abacus.carry[w->abacus.state] > 0) ? "carry" : "borrow"),
			      ABS(w->abacus.carry[w->abacus.state]));
			}
		}
		(void) sprintf(buffer2, "%s.", buffer2);
		drawLineText(w, buffer2, 1);
		if (w->abacus.carry[w->abacus.state] == 0 &&
				w->abacus.carryStep == 0 && done) {
			drawLineText(w, "Done.", 0);
			w->abacus.step = 0;
		} else if (w->abacus.carryStep == 0) {
			w->abacus.step++;
		} else {
			w->abacus.carryStep++;
		}
	} else {
		int rPosition = 0;
		int aDigit = 0, bDigit = 0;
		Boolean done = False;

		if (!w->abacus.rightToLeftAdd)
			w->abacus.carry[1] = w->abacus.carry[0];
		w->abacus.state = 1;
		if (w->abacus.op == '+' || w->abacus.op == '-') {
			done = nextPositionSum(w, w->abacus.op,
				&rPosition, &aDigit, &bDigit);
			setStringBuffer(w, w->abacus.rString, w->abacus.op,
				rPosition, w->abacus.lower, w->abacus.upper);
		} else if (w->abacus.op == '*') {
			done = nextPositionProduct(w,
				&rPosition, &aDigit, &bDigit);
			setStringBuffer(w, w->abacus.rString, '+',
				rPosition, w->abacus.lower, w->abacus.upper);
		} else if (w->abacus.op == '/') {
			done = nextPositionDivision(w,
				&rPosition, &aDigit, &bDigit);
			setStringBuffer(w, w->abacus.rString, '-',
				rPosition, w->abacus.lower, w->abacus.upper);
		}
		if (w->abacus.lower != 0)
			setAbacusMove(w, ACTION_MOVE, 0, 0, rPosition,
				w->abacus.lower);
		if (w->abacus.upper != 0)
			setAbacusMove(w, ACTION_MOVE, 0, 1, rPosition,
				w->abacus.upper);
#if 0
		if (w->abacus.carry[w->abacus.state] != 0) {
			setAbacusMove(w, ACTION_MOVE, 0, 0, rPosition + 1,
				w->abacus.carry[w->abacus.state]);
			w->abacus.carry[w->abacus.state] = 0;
		}
#endif
		if (w->abacus.carry[w->abacus.state] == 0 &&
				w->abacus.carryStep != 0) {
			w->abacus.carryStep = 0;
		}
		if (w->abacus.carry[w->abacus.state] == 0 &&
				w->abacus.carryStep == 0 &&
				done) {
			drawLineText(w, "Done.", 0);
			w->abacus.step = 0;
		} else if ((done && w->abacus.rightToLeftAdd) ||
                                        !w->abacus.rightToLeftAdd) {
			if (w->abacus.carry[w->abacus.state] != 0) {
				if (w->abacus.carryStep == 0) {
					w->abacus.carryStep = 2;
					if (w->abacus.rightToLeftAdd)
						w->abacus.carry[1] = w->abacus.carry[0] = 0;
				} else {
					w->abacus.carryStep++;
				}
			}
			if (w->abacus.carryStep == 0)
				w->abacus.step++;
		} else if (w->abacus.rightToLeftAdd) {
			w->abacus.step++;
		}
	}
}
